# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
''' Contexts for *with* statement providing temporary directories
'''
from __future__ import with_statement
import os
import shutil
from tempfile import template, mkdtemp


class TemporaryDirectory(object):
    """Create and return a temporary directory.  This has the same
    behavior as mkdtemp but can be used as a context manager.

    Upon exiting the context, the directory and everthing contained
    in it are removed.

    Examples
    --------
    >>> import os
    >>> with TemporaryDirectory() as tmpdir:
    ...     fname = os.path.join(tmpdir, 'example_file.txt')
    ...     with open(fname, 'wt') as fobj:
    ...         _ = fobj.write('a string\\n')
    >>> os.path.exists(tmpdir)
    False
    """
    def __init__(self, suffix="", prefix=template, dir=None, chdir=False):
        self.name = mkdtemp(suffix, prefix, dir)
        self._closed = False

    def __enter__(self):
        return self.name

    def cleanup(self):
        if not self._closed:
            shutil.rmtree(self.name)
            self._closed = True

    def __exit__(self, exc, value, tb):
        self.cleanup()
        return False


class InTemporaryDirectory(TemporaryDirectory):
    ''' Create, return, and change directory to a temporary directory

    Examples
    --------
    >>> import os
    >>> my_cwd = os.getcwd()
    >>> with InTemporaryDirectory() as tmpdir:
    ...     _ = open('test.txt', 'wt').write('some text')
    ...     assert os.path.isfile('test.txt')
    ...     assert os.path.isfile(os.path.join(tmpdir, 'test.txt'))
    >>> os.path.exists(tmpdir)
    False
    >>> os.getcwd() == my_cwd
    True
    '''
    def __enter__(self):
        self._pwd = os.getcwd()
        os.chdir(self.name)
        return super(InTemporaryDirectory, self).__enter__()

    def __exit__(self, exc, value, tb):
        os.chdir(self._pwd)
        return super(InTemporaryDirectory, self).__exit__(exc, value, tb)



class InGivenDirectory(object):
    """ Clean out, change directory to given directory

    Useful for debugging
    """
    def __init__(self, path):
        self.path = os.path.abspath(path)

    def __enter__(self):
        self._pwd = os.path.abspath(os.getcwd())
        if os.path.isdir(self.path):
            shutil.rmtree(self.path)
        os.mkdir(self.path)
        os.chdir(self.path)
        return self.path

    def __exit__(self, exc, value, tb):
        os.chdir(self._pwd)



import os
from os.path import join as pjoin, abspath
import grp
import pwd
import sys
from distutils.util import spawn
from distutils.version import StrictVersion, LooseVersion
from distutils.dir_util import mkpath
from distutils.file_util import copy_file
import distutils.core
from tempfile import mkdtemp

from .tmpdirs import InTemporaryDirectory
from .py3k import unicode


def Version(s):
    try:
        return StrictVersion(s)
    except ValueError:
        return LooseVersion(s)

def run_setup(*args, **kwargs):
    """
    Re-entrant version of distutils.core.run_setup()
    """
    PRESERVE = '_setup_stop_after', '_setup_distribution'
    d = {}
    for k in PRESERVE:
        try:
            d[k] = getattr(distutils.core, k)
        except AttributeError:
            pass
    try:
        return distutils.core.run_setup(*args, **kwargs)
    finally:
        for k,v in d.items():
            setattr(distutils.core, k, v)

def adminperms(src, verbose=0, dry_run=0):
    try:
        spawn(['/usr/bin/chgrp', '-R', 'admin', src])
        spawn(['/bin/chmod', '-R', 'g+w', src])
    except:
        return False
    return True

def mkbom(src, pkgdir, verbose=0, dry_run=0, TOOL='/usr/bin/mkbom'):
    """
    Create a bill-of-materials (BOM) for the given src directory and store it
    to the given pkg directory
    """
    dest = os.path.join(pkgdir, 'Contents', 'Archive.bom')
    mkpath(os.path.dirname(dest), verbose=verbose, dry_run=dry_run)
    spawn([TOOL, src, dest], verbose=verbose, dry_run=dry_run)

def pax(src, pkgdir, verbose=0, dry_run=0, TOOL='/bin/pax'):
    """
    Create a pax gzipped cpio archive of the given src directory and store it
    to the given pkg directory

    returns size of archive
    """
    dest = os.path.realpath(os.path.join(pkgdir, 'Contents', 'Archive.pax.gz'))
    mkpath(os.path.dirname(dest), verbose=verbose, dry_run=dry_run)
    pwd = os.path.realpath(os.getcwd())
    os.chdir(src)
    try:
        spawn([TOOL, '-w', '-f', dest, '-x', 'cpio', '-z', '.'])
    finally:
        os.chdir(pwd)
    return os.stat(dest).st_size


def find_paxboms(base_path):
    """ Generator returning pax / bom pairs from root `base_path`

    Parameters
    ----------
    base_path : str
        directory from which to search recursively for pax / bom pairs

    Yields
    ------
    px_bom : length 2 tuple
        tuple of (pax file, bom file)
    """
    for root, dirs, files in os.walk(base_path):
        for file in files:
            if not file.endswith('.bom'):
                continue
            froot = file[:-3]
            for end in ('pax', 'pax.gz'):
                if froot + end in files:
                    yield (pjoin(root, froot + end),
                           pjoin(root, file))


def ugrp_path(path):
    """ Return user and group for a path `path`
    """
    stat_info = os.stat(path)
    uid = stat_info.st_uid
    gid = stat_info.st_gid
    user = pwd.getpwuid(uid)[0]
    group = grp.getgrgid(gid)[0]
    return user, group


def unpax(pax_file, out_dir=None, TOOL='/bin/pax'):
    """ Unpack a pax archive `pax_file` into `out_dir`

    Parameters
    ----------
    pax_file : str
        filename of pax file
    out_dir : None or str, optional
        directory into which to unpack files; if None, create temporary
        directory and unpack there.  The caller should delete the directory.
    TOOL : str, optional
        path to ``pax`` binary

    Returns
    -------
    out_dir : str
        directory containing unpacked files
    """
    pwd = os.path.realpath(os.getcwd())
    pax_path = os.path.abspath(pax_file)
    cmd = [TOOL, '-r', '-p', 'e', '-f', pax_path]
    if pax_file.endswith('.gz'):
        cmd += ['-z']
    elif pax_file.endswith('.bz2'):
        cmd += ['-j']
    if out_dir is None:
        out_dir = mkdtemp
    os.chdir(out_dir)
    try:
        spawn(cmd)
    finally:
        os.chdir(pwd)
    return out_dir


def reown_paxboms(base_path, user, group, TOOL='/usr/sbin/chown'):
    """ Change ownnerhsip files in pax/boms within `base_path`

    Parameters
    ----------
    base_path : str
        path to tree somewhere containing bom / pax pairs. We will change
        ownership of files within the pax archive, and record this in the bom.
    user : str
        user to which to change ownership
    group : str
        group to which to change ownership
    """
    for pxbom in find_paxboms(base_path):
        px, bm = [abspath(f) for f in pxbom]
        with InTemporaryDirectory() as tmpdir:
            arch_path = pjoin(tmpdir, 'archive')
            os.mkdir(arch_path)
            unpax(px, arch_path)
            spawn([TOOL, '-R', '%s:%s' % (user, group), arch_path])
            os.mkdir('Contents')
            pax(arch_path, tmpdir)
            mkbom(arch_path, tmpdir)
            rs1 = copy_file(pjoin('Contents', 'Archive.bom'), bm)
            rs2 = copy_file(pjoin('Contents', 'Archive.pax.gz'), px)
            assert rs1 == (bm, True)
            assert rs2 == (px, True)


def unicode_path(path, encoding=sys.getfilesystemencoding()):
    if isinstance(path, unicode):
        return path
    return unicode(path, encoding)

def walk_files(path):
    for root, dirs, files in os.walk(path):
        for fn in files:
            yield os.path.join(root, fn)

def get_gid(name):
    """ Return integer gid for group with name `name`
    """
    return grp.getgrnam(name).gr_gid

def find_root(path, base='/'):
    """
    Return the list of files, the archive directory, and the destination path
    """
    files = list(walk_files(path))
    common = os.path.dirname(os.path.commonprefix(files))
    prefix = os.path.join(base, common[len(os.path.join(path, '')):])
    #while not os.path.exists(prefix):
    #    common = os.path.dirname(common)
    #    prefix = os.path.dirname(prefix)
    prefix = os.path.realpath(prefix)
    return files, common, prefix

def admin_writable(path):
    gid = get_gid('admin')
    while not os.path.exists(path):
        path = os.path.dirname(path)
    s = os.stat(path)
    mode = s.st_mode
    return (mode & 0x2) or (s.st_gid == gid and mode & 0x10)

def reduce_size(files):
    return sum([os.stat(fn).st_size for fn in files])

def sw_vers(_cache=[]):
    if not _cache:
        info = os.popen('/usr/bin/sw_vers').read().splitlines()
        for line in info:
            key, value = line.split(None, 1)
            if key == 'ProductVersion:':
                _cache.append(Version(value.strip()))
                break
        else:
            raise ValueError("sw_vers not behaving correctly")
    return _cache[0]

def is_framework_python():
    return os.path.dirname(os.path.dirname(sys.prefix)).endswith('.framework')



""" Test some tools
"""
from __future__ import with_statement

import os
from os.path import abspath, split as psplit, isfile, join as pjoin

from ..tools import pax, unpax, ugrp_path

from ..tmpdirs import InTemporaryDirectory

from nose.tools import assert_true, assert_equal

HERE, SELF = psplit(abspath(__file__))

def test_pax_unpax():
    with InTemporaryDirectory() as tmpdir:
        pax(HERE, tmpdir)
        pax_file = pjoin(tmpdir, 'Contents', 'Archive.pax.gz')
        assert_true(isfile(pax_file))
        os.mkdir('pax_contents')
        unpax(pax_file, 'pax_contents')
        assert_equal(set(os.listdir('pax_contents')),
                     set(os.listdir(HERE)))


def test_user_group():
    with InTemporaryDirectory() as tmpdir:
        open('test_file', 'wt').write('hello')
        assert_equal(ugrp_path(tmpdir), ugrp_path('test_file'))



""" Extraordinarily crude run-oneself test

You'll need ``nose`` installed to run this.

Run all tests with ``nosetests bdist_mpkg`` from the root directory (containing
the ``setup.py`` file).
"""
from __future__ import with_statement

import sys
import os
import pwd
from os.path import dirname, join as pjoin, isfile, isdir
from subprocess import check_call
from shutil import rmtree

from ..tools import reown_paxboms, find_paxboms, unpax, ugrp_path

from ..tmpdirs import TemporaryDirectory

from nose import SkipTest
from nose.tools import assert_true, assert_equal

MY_PATH = dirname(__file__)
MY_PYTHON = sys.executable

def test_myself():
    # Build myself into a temporary directory
    os.chdir(pjoin(MY_PATH, '..', '..'))
    if not isfile('setup.py'):
        raise SkipTest('Not running from development directory')
    with TemporaryDirectory() as tmpdir:
        cmd = '%s setup.py bdist_mpkg --dist-dir=%s' % (MY_PYTHON, tmpdir)
        check_call(cmd, shell=True)
        tmpls = os.listdir(tmpdir)
        assert_equal(len(tmpls), 1)
        mpkg_dir = tmpls[0]
        assert_true(mpkg_dir.endswith('.mpkg'))
        assert_true(isdir(pjoin(tmpdir, mpkg_dir)))
        # Check zipping
        cmd += ' -z'
        check_call(cmd, shell=True)
        tmpls = sorted(os.listdir(tmpdir))
        assert_equal(len(tmpls), 2)
        assert_equal(tmpls[0], mpkg_dir)
        assert_true(tmpls[1].endswith('.zip'))
        assert_true(isfile(pjoin(tmpdir, tmpls[1])))
        # Try reperming
        mpkg_dir = pjoin(tmpdir, mpkg_dir)
        # Find paxboms
        paxboms = list(find_paxboms(mpkg_dir))
        assert_equal(len(paxboms), 2) # scripts and library
        # Extract them and check permissions
        paxdir = pjoin(tmpdir, 'pax_files')
        grps = []
        for pxbom in paxboms:
            rmtree(paxdir, ignore_errors=True)
            os.mkdir(paxdir)
            unpax(pxbom[0], paxdir)
            user, grp = ugrp_path(pjoin(paxdir, os.listdir(paxdir)[0]))
            assert_equal(user, pwd.getpwuid(os.getuid())[0])
            grps.append(grp)
        assert_equal([g for g in grps if g != grp], [])
        # Do paxbom permission change
        reown_paxboms(mpkg_dir, user, "everyone")
        for pxbom in paxboms:
            rmtree(paxdir, ignore_errors=True)
            os.mkdir(paxdir)
            unpax(pxbom[0], paxdir)
            user, grp = ugrp_path(pjoin(paxdir, os.listdir(paxdir)[0]))
            assert_equal(user, pwd.getpwuid(os.getuid())[0])
            assert_equal(grp, 'everyone')



import os
import sys
from plistlib import Plist

from . import __version__ as bdist_mpkg_version
from . import tools
from .py3k import unicode, u, any_str_type

def _major_minor(v):
    rval = [0, 0]
    try:
        for i, rev in enumerate(v.version):
            rval[i] = int(rev)
    except (TypeError, ValueError, IndexError):
        pass
    return rval

def common_info(name, version):
    # Keys that can appear in any package
    name, version = unicode(name), tools.Version(version)
    major, minor = _major_minor(version)
    return dict(
        CFBundleGetInfoString=u('%s %s' % (name, version)),
        CFBundleIdentifier=u('org.pythonmac.%s' % (name,)),
        CFBundleName=name,
        CFBundleShortVersionString=unicode(version),
        IFMajorVersion=major,
        IFMinorRevision=minor,
        IFPkgFormatVersion=0.10000000149011612,
        IFRequirementDicts=[path_requirement(u('/'))],
        PythonInfoDict=dict(
            PythonLongVersion=unicode(sys.version),
            PythonShortVersion=unicode(sys.version[:3]),
            PythonExecutable=unicode(sys.executable),
            bdist_mpkg=dict(
                version=unicode(bdist_mpkg_version),
            ),
        ),
    )

def pkg_info(name, version):
    d = common_info(name, version)
    # Keys that can only appear in single packages
    d.update(dict(
        IFPkgFlagAllowBackRev=False,
        IFPkgFlagAuthorizationAction=u('AdminAuthorization'),
        #IFPkgFlagDefaultLocation=u'/Library/Python/2.3',
        IFPkgFlagFollowLinks=True,
        IFPkgFlagInstallFat=False,
        IFPkgFlagIsRequired=False,
        IFPkgFlagOverwritePermissions=False,
        IFPkgFlagRelocatable=False,
        IFPkgFlagRestartAction=u('NoRestart'),
        IFPkgFlagRootVolumeOnly=True,
        IFPkgFlagUpdateInstalledLangauges=False,
    ))
    return d

def path_requirement(SpecArgument, Level=u('requires'), **kw):
    return dict(
        Level=Level,
        SpecType=u('file'),
        SpecArgument=tools.unicode_path(SpecArgument),
        SpecProperty=u('NSFileType'),
        TestOperator=u('eq'),
        TestObject=u('NSFileTypeDirectory'),
        **kw
    )

FRIENDLY_PREFIX = {
    os.path.expanduser(u('~/Library/Frameworks')) : u('User'),
    u('/System/Library/Frameworks') : u('Apple'),
    u('/Library/Frameworks') : u('python.org'),
    u('/opt/local') : u('DarwinPorts'),
    u('/usr/local') : u('Unix'),
    u('/sw') : u('Fink'),
}

def python_requirement(pkgname, prefix=None, version=None, **kw):
    if prefix is None:
        prefix = sys.prefix
    if version is None:
        version = sys.version[:3]
    prefix = os.path.realpath(prefix)
    fmwkprefix = os.path.dirname(os.path.dirname(prefix))
    is_framework = fmwkprefix.endswith('.framework')
    if is_framework:
        dprefix = os.path.dirname(fmwkprefix)
    else:
        dprefix = prefix
    dprefix = tools.unicode_path(dprefix)
    name = u('%s Python %s' % (FRIENDLY_PREFIX.get(dprefix, dprefix), version))
    kw.setdefault('LabelKey', name)
    title = u('%s requires %s to install.' % (pkgname, name,))
    kw.setdefault('TitleKey', title)
    kw.setdefault('MessageKey', title)
    return path_requirement(prefix, **kw)

def mpkg_info(name, version, packages=[]):
    d = common_info(name, version)
    # Keys that can appear only in metapackages
    npackages = []
    for p in packages:
        items = getattr(p, 'items', None)
        if items is not None:
            p = dict(items())
        else:
            if isinstance(p, any_str_type):
                p = [p]
            p = dict(zip(
                (u('IFPkgFlagPackageLocation'), u('IFPkgFlagPackageSelection')),
                p
            ))
        npackages.append(p)
    d.update(dict(
        IFPkgFlagComponentDirectory=u('./Contents/Packages'),
        IFPkgFlagPackageList=npackages,
    ))
    return d

def checkpath_plugin(path):
    # This does not appear to be called in this codebase
    if not isinstance(path, unicode):
        path = unicode(path, encoding) # encoding not defined
    return dict(
        searchPlugin=u('CheckPath'),
        path=path,
    )

def common_description(name, version):
    name, version = unicode(name), tools.Version(version)
    return dict(
        IFPkgDescriptionTitle=name,
        IFPkgDescriptionVersion=unicode(version),
    )

def write(dct, path):
    p = Plist()
    p.update(dct)
    p.write(path)



#!/usr/bin/env python
""" Script requires interactive sudo password

Checks that installer installs
"""

import sys
from os import chdir
from os.path import isfile, isdir, dirname
from glob import glob
from distutils import sysconfig

from subprocess import check_call
from functools import partial

from bdist_mpkg.tools import is_framework_python

from nose import SkipTest
from nose.tools import assert_true, assert_false, assert_equal

call = partial(check_call, shell=True)

VOL_NAME = 'check_install'
IMG_PATH = '/Volumes/' + VOL_NAME

MYDIR = dirname(__file__)
chdir(MYDIR + '/../..')
if not isfile('setup.py'):
    raise SkipTest(
""" These tests run from the development repository, or an unpacked sdist.
They won't run from an installed copy""")


def install2dmg(dmg_path, opts_str=''):
    # Make, get mpkg installer in dist
    call('python setup.py bdist_mpkg ' + opts_str)
    post_mpkgs = glob('dist/*.mpkg')
    assert_equal(len(post_mpkgs), 1)
    new_mpkg = post_mpkgs.pop()
    # Make disk image to install to
    call('hdiutil create -size 5m -fs HFS+ -volname %s %s'
        % (VOL_NAME, dmg_path))
    # Mount disk image
    call('hdiutil mount ' + dmg_path)
    # Install
    call('sudo installer -pkg %s -target %s' %
        (new_mpkg, IMG_PATH))

# Nuclear clean
call('git clean -fxd')
# Build mpkg and install to dmg
install2dmg('build/check_install.dmg')
# Check
assert_true(isdir("%s%s/bdist_mpkg"
                  % (IMG_PATH, sysconfig.get_python_lib())))
if is_framework_python():
    assert_true(isfile("%s/usr/local/bin/bdist_mpkg" % (IMG_PATH,)))
else:
    assert_true(isfile("%s%s/bin/bdist_mpkg" % (IMG_PATH, sys.exec_prefix)))
call('hdiutil unmount ' + IMG_PATH)
# Check not frameworking
install2dmg('build/check_nf_install.dmg', '--local-scripts')
# Check
assert_true(isdir("%s%s/bdist_mpkg"
                  % (IMG_PATH, sysconfig.get_python_lib())))
assert_false(isfile("%s/usr/local/bin/bdist_mpkg" % (IMG_PATH,)))
assert_true(isfile("%s%s/bin/bdist_mpkg" % (IMG_PATH, sys.exec_prefix)))
call('hdiutil unmount ' + IMG_PATH)



#!/usr/bin/env python

from setuptools import setup

import os

# BEFORE importing distutils, remove MANIFEST. distutils doesn't properly
# update it when the contents of directories change.
if os.path.exists('MANIFEST'): os.remove('MANIFEST')

# Get version and release info, which is all stored in bdist_mpkg/info.py
ver_file = os.path.join('bdist_mpkg', 'info.py')
VARS = {}
exec(open(ver_file).read(), VARS)

setup(
    name="bdist_mpkg",
    version=VARS['__version__'],
    description=VARS['DESCRIPTION'],
    long_description=VARS['LONG_DESCRIPTION'],
    classifiers=VARS['CLASSIFIERS'],
    author="Bob Ippolito",
    author_email="pythonmac-sig@python.org",
    url="https://github.com/matthew-brett/bdist_mpkg",
    license="MIT License",
    packages=['bdist_mpkg'],
    platforms=['any'],
    zip_safe=True,
    entry_points={
        'distutils.commands': [
            'bdist_mpkg = bdist_mpkg.cmd_bdist_mpkg:bdist_mpkg',
        ],
        'console_scripts': [
            'bdist_mpkg = bdist_mpkg.script_bdist_mpkg:main',
            'reown_mpkg = bdist_mpkg.script_reown_mpkg:main',
        ],
    },
)



""" This file contains defines parameters for bdist_mpkg that we use to fill
settings in setup.py, the bdist_mpkg top-level docstring, and for building the
docs.  In setup.py in particular, we exec this file, so it cannot import
bdist_mpkg """

# bdist_mpkg version information.  An empty _version_extra corresponds to a
# full release.  '.dev' as a _version_extra string means this is a development
# version
_version_major = 0
_version_minor = 5
_version_micro = 0
#_version_extra = 'dev'
_version_extra = ''

# Format expected by setup.py and doc/source/conf.py: string of form "X.Y.Z"
__version__ = "%s.%s.%s%s" % (_version_major,
                              _version_minor,
                              _version_micro,
                              _version_extra)

DESCRIPTION = "Builds Mac OS X installer packages from distutils"
LONG_DESCRIPTION = """
==========
bdist_mpkg
==========

bdist_mpkg is a distutils plugin that implements the ``bdist_mpkg`` command.
The command builds a Mac OS X metapackage for use by Installer.app for easy GUI
installation of Python modules, much like ``bdist_wininst``.

It also comes with a ``bdist_mpkg`` script, which is a setup.py front-end that
will allow you to easy build an installer metapackage from nearly any existing
package that uses distutils.

Please email the `Python-Mag SIG mailing list
<http://www.python.org/community/sigs/current/pythonmac-sig/>`_ with questions,
and let us know of bugs via `github issues
<https://github.com/matthew-brett/bdist_mpkg/issues>`_

Code
====

The code started life at:

http://undefined.org/python/#bdist_mpkg

Bob Ippolito wrote most of the code.

The `current repository`_ is on Github.

.. _current repository: http://github.com/matthew-brett/bdist_mpkg

License
=======

MIT license.  See the ``LICENSE`` file in the source archive.
"""

CLASSIFIERS = filter(None, map(str.strip,
"""
Intended Audience :: Developers
License :: OSI Approved :: MIT License
Programming Language :: Python
Operating System :: MacOS :: MacOS X
Topic :: Software Development :: Libraries :: Python Modules
Topic :: Software Development :: Build Tools
""".splitlines()))



#!/usr/bin/env python
DESCRIP = 'Change ownership of files in mpkg archives'
EPILOG = \
"""Unpack archive files, change ownership, pack again

This may be useful for post-processing an mpkg directory.

For example, say you wanted to make the files in the directory belong to root;
then you could first build the package as non-root (say on buildbots), and then
fetch the package to a safe machine, and do::

    sudo reown_mpkg path_to_built.mpkg root admin

to fix up the permissions.
"""
from os.path import isdir

from argparse import (ArgumentParser, RawDescriptionHelpFormatter)

from .tools import reown_paxboms

def main():
    parser = ArgumentParser(description=DESCRIP,
                            epilog=EPILOG,
                            formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument('mpkg_path', type=str,
                        help='mpkg directory')
    parser.add_argument('user', type=str,
                        help='username to become owner')
    parser.add_argument('group', type=str,
                        help='group to become owner')
    args = parser.parse_args()
    if not isdir(args.mpkg_path):
        raise RuntimeError('No directory "%s"' % args.mpkg_path)
    reown_paxboms(args.mpkg_path, args.user, args.group)


if __name__ == '__main__':
    main()
